SUBROUTINE comb(n, k, comba, c)
!#############################################################################
! comb returns an array with all combinations of k numbers in n positions
!
! author: Florian Boecker   date: 25/10/2011
! http://complex-systems.uni-muenster.de/
! Licensing:  This code is distributed under the GNU GPL license.
!
! parameters:
!   name      i/o     datatype   description
!   --------------------------------------------------------------------------
!   n:        input   integer    - the total number of things
!   k:        input   integer    - the number of things in each combination
!   c:        input   integer    - the number of combinations to be generated
!   comba:    output  integer    - the returned array with the combinations,
!                                  should contain at least the c rows and
!                                  k columns
!#############################################################################
    IMPLICIT NONE
    INTEGER n, k, i, c, a(k)
    INTEGER, DIMENSION(c, k) :: comba
    LOGICAL done
    done = .TRUE.
    i = 1
    DO i = 1, c
      CALL comb_next ( n, k, a, done )
      comba(i,:) = a(:)
      IF(done) THEN
        EXIT
      END IF
    END DO
END SUBROUTINE comb

subroutine comb_next ( n, k, a, done )

!*****************************************************************************80
!
!! COMB_NEXT computes combinations of K things out of N.
!
!  Discussion:
!
!    The combinations are computed one at a time, in lexicographical order.
!
!    10 April 1009: Thanks to "edA-qa mort-ora-y" for supplying a
!    correction to this code!
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    10 April 2009
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Charles Mifsud,
!    Algorithm 154:
!    Combination in Lexicographic Order,
!    Communications of the ACM,
!    March 1963.
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) N, the total number of things.
!
!    Input, integer ( kind = 4 ) K, the number of things in each combination.
!
!    Input/output, integer ( kind = 4 ) A(K), contains the list of elements in
!    the current combination.
!
!    Input/output, logical DONE.  On first call, set DONE to TRUE,
!    and thereafter, its input value should be the output value from
!    the previous call.  The output value will normally be FALSE,
!    indicating that there are further combinations that can be
!    returned.  When DONE is returned TRUE, the sequence is exhausted.
!
  implicit none

  integer ( kind = 4 ) k

  integer ( kind = 4 ) a(k)
  logical              done
  integer ( kind = 4 ) i
  integer ( kind = 4 ) j
  integer ( kind = 4 ) n

  if ( done ) then

    if ( k <= 0 ) then
      return
    end if

    call i4vec_indicator ( k, a )

    done = .false.

  else

    if ( a(k) < n ) then
      a(k) = a(k) + 1
      return
    end if

    do i = k, 2, -1

      if ( a(i-1) < n-k+i-1 ) then

        a(i-1) = a(i-1) + 1

        do j = i, k
          a(j) = a(i-1) + j - ( i-1 )
        end do

        return

      end if

    end do

    done = .true.

  end if

  return
end

subroutine i4vec_indicator ( k, a )

!*****************************************************************************80
!
!! I4VEC_INDICATOR sets an I4VEC to the indicator vector.
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    09 November 2000
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) k, the number of elements of A.
!
!    Output, integer ( kind = 4 ) A(k), the array to be initialized.
!
  implicit none

  integer ( kind = 4 ) k

  integer ( kind = 4 ) a(k)
  integer ( kind = 4 ) i

  do i = 1, k
    a(i) = i
  end do

  return
end
