SUBROUTINE map_tma_data_file(datafilename, mapfilename, ncol, nrow, nfactor, nrecord, mapping, n)
!#############################################################################################
! maps a csv-tma-data-file according to a mapping file
!
! author: Florian Boecker   date: 19/04/2012
! http://complex-systems.uni-muenster.de/
! Licensing:  This code is distributed under the GNU GPL license.
!
! parameters:
!   name            i/o     datatype   description
!   -------------------------------------------------------------------------------------------
!   datafilename:   input   character       - the complete path of the file to be read in
!   mapfilename:    input   character       - the complete path of the file with
!                                             the mapping of the factors
!   ncol:           output  intger          - number of columns in the data file
!   nrow:           output  integer         - number of row in the data file
!   nfactor:        output  integer         - number of factors in the data
!   nrecord:        output  integer         - number of data records
!   mapping:        output  integer(100)    - mapping of the factors
!   n:              output  integer         - number of factors to be incoperated and also the
!                                             length of the mapping
!###############################################################################################
    IMPLICIT  NONE
    ! Declaration
    CHARACTER (len=200) :: datafilename ! paths and name of the data file
    CHARACTER (len=200) :: mapfilename  ! paths and name of the mapping file
    CHARACTER (len=200) :: record       ! csv-file record buffer
    CHARACTER csv_char_header           ! holds 'y' or 'n' indicating whether the data file is expected to have a header
    CHARACTER csv_char_linenames        ! holds 'y' or 'n' indicating whether the data file is expected to have a line number in the first column
    INTEGER value_count                 ! number of values in a tab separated record
    INTEGER mapfileunit                 ! ID of the map-file
    INTEGER datafileunit                ! ID of the tma-data-file
    INTEGER ncol                        ! number of columns in the data file
    INTEGER nrow                        ! number of lines in the data file
    INTEGER nfactor                     ! number of factors in the data
    INTEGER nrecord                     ! number of data records
    INTEGER, DIMENSION(100) :: mapping  ! mapping of the factors
    INTEGER n                           ! number of factors to be incorporated and also the length of the mapping
    INTEGER input_status                ! READ error handling
    INTEGER maximum                     ! highest factor number found in the mapping

    ! Initialisation
    mapping(:) = 0

    ! Determine data file dimensions
    CALL csvmeasurements(datafilename, nrow, ncol)

    ! Read mapping file
    CALL get_unit(mapfileunit)
    OPEN(unit = mapfileunit, file = mapfilename)
    READ(mapfileunit,'(A)', iostat = input_status) record
    CLOSE(mapfileunit)
    IF(input_status /= 0) THEN
        WRITE(*,*) 'Error in mapping file'
        WRITE(*,*) 'Programm canceled'
        STOP
    ENDIF

    ! Determine number of factors to be incorporated and the length of the mapping
    CALL csv_value_count(record, value_count)
    IF(value_count < 2) THEN
        WRITE(*,*) 'Error in mapping file:'
        WRITE(*,*) 'Mapping is supposed to be tab-separated, but no tabs were found'
        WRITE(*,*) 'Programm canceled'
        STOP
    ENDIF
    n = value_count-2

    ! Reading record
    READ(record, *, iostat = input_status) csv_char_header, csv_char_linenames, mapping(1:n)
    IF(input_status /= 0) THEN
        WRITE(*,*) 'Error in mapping file'
        WRITE(*,*) 'Programm canceled'
        STOP
    ENDIF

    ! Determine whether header is to be expected
    IF((csv_char_header == 'Y') .OR. (csv_char_header == 'y')) THEN
        nrecord = nrow-1
    ELSE
        IF((csv_char_header == 'N') .OR. (csv_char_header == 'n')) THEN
            nrecord = nrow
        ELSE
            WRITE(*,*) 'Error in mapping file: Illegal header specification'
            WRITE(*,*) 'Programm canceled'
            STOP
        ENDIF
    ENDIF

    ! Determine whether line names are to be expected
    IF((csv_char_linenames == 'Y') .OR. (csv_char_linenames == 'y')) THEN
        nfactor = ncol-1
    ELSE
        IF((csv_char_linenames == 'N') .OR. (csv_char_linenames == 'n')) THEN
            nfactor = ncol
        ELSE
            WRITE(*,*) 'Error in mapping file: Illegal line names specification'
            WRITE(*,*) 'Programm canceled'
            STOP
        ENDIF
    ENDIF

    ! Check wether mapping fits the data file
    IF((n > nfactor) .OR. (MAXVAL(mapping(:)) > nfactor)) THEN
        WRITE(*,*) 'Mapping does not fit the data file'
        WRITE(*,*) 'Programm canceled'
        STOP
    ENDIF

END SUBROUTINE map_tma_data_file

SUBROUTINE csvmeasurements ( filename, nline, ncol )
  IMPLICIT NONE

  character ( len = * )    filename
  integer   ( kind = 4 )   ierror
  integer   ( kind = 4 )   input_status
  integer   ( kind = 4 )   input_unit
  character ( len = 200 )  line
  integer   ( kind = 4 )   nline
  integer   (kind = 4)     ncol

  call get_unit ( input_unit )

  open ( unit = input_unit, file = filename)

  nline = 0

  do
    read ( input_unit, '(a)', iostat = input_status ) line
    IF (nline == 0) THEN
        CALL csv_value_count (line, ncol)
    ENDIF

    if ( input_status /= 0 ) then
      ierror = nline
      exit
    end if

    nline = nline + 1

  end do

  close ( unit = input_unit )

  return
END SUBROUTINE csvmeasurements

subroutine csv_value_count ( csv_record, value_count )
    implicit none
    character              csv_char
    character              csv_char_old
    integer   ( kind = 4 ) csv_len
    integer   ( kind = 4 ) csv_loc
    character ( len  = * ) csv_record
    integer   ( kind = 4 ) value_count
    !  No values so far.
    value_count = 1
    !  How many characters in the record?
    csv_len = len_trim ( csv_record )
    !  Consider each character.
    DO csv_loc = 1, csv_len
        csv_char = csv_record(csv_loc:csv_loc)
        !  Each tab divides one value from another.
        if ( csv_char == CHAR(9) ) then ! CHAR(9) == Tab
            value_count = value_count + 1
        END IF
    END DO
    RETURN
END SUBROUTINE csv_value_count

