SUBROUTINE perm_tj(data, nref, ntest, ssq_best, ssq_g, test_order)
    !prevent implicite variable declaration
    IMPLICIT NONE

    !#############################################################################
    ! Perm_tj returns the minimal global SSQ for all test set permutations,
    ! the optimal order of the test factors, and for each reference
    ! factor the SSQ for the optimal rank order.
    !
    ! Uses the Trotter-Johnson permutation algorithm.
    !
    ! author: Florian Boecker   date: 05/01/2012
    ! http://complex-systems.uni-muenster.de/
    ! Licensing:  This code is distributed under the GNU GPL license.
    !
    ! parameters:
    !   name        i/o     datatype      description
    !   ----------------------------------------------------------------------------
    !   data        input   real(:,:)     - the correlation matrix
    !   nref        input   integer       - number of factors in the reference set
    !   ntest       input   integer       - number of factors in the test set
    !   ssq_best    output  real(:)       - the ssqs for every reference factor
    !                                       resulting in the optimal ssq
    !   ssq_g       output  real          - sum of ssq_best: the lowest and ergo
    !                                       optimal ssq
    !   test_order  output  integer(:)    - the order of test factors that resulted
    !                                       in the optimal ssq
    !################################################################################

    !declaration
    !   parameters
    REAL, DIMENSION(ntest, nref) :: data !the correlation matrix (input)
    INTEGER nref        !number of factors in the reference set (input)
    INTEGER ntest       !number of factors in the test set (input)
    REAL, DIMENSION(nref) :: ssq_best    !vector with the ssqs for every reference factor resulting in the optimal ssq (output)
    REAL ssq_g       !sum of ssq_best: the lowest and ergo optimal ssq (output)
    INTEGER, DIMENSION(ntest) :: test_order !the order of test factors that resulted in the optimal ssq (output)
    !   varibales for internal use
    INTEGER rank    !rank of the permuation
    INTEGER, DIMENSION(ntest) :: p  !permutation
    REAL, DIMENSION(nref) :: ssq_s !vector with the ssq for every reference factor
    REAL ssq    ! the sum of squares of a permutation
    INTEGER i, j   !loop control
    REAL, DIMENSION(ntest) ::  part_cor !set of correlation values picked from the correlation matrix according to the permutation
    INTEGER (kind = 8) maxrank !rank that terminates the permutation
    REAL, DIMENSION(ntest) ::  x
    !report
    !INTEGER iunit   ! unitID for report file
    !CHARACTER (len = 200 ) record ! buffer string for output

    !initialisation
    rank = -1   !tells perm_tj_successor to create the first permuation (rank = 0) (1,2,3,...,ntest)
    ssq_g = huge(1)
    maxrank = 1
    DO i = 1, ntest
        maxrank = maxrank * i
        x(i) = real(i)
    END DO
    maxrank = maxrank/2

    !calculation
    DO
        CALL perm_tj_successor(ntest, p, rank)
        IF(rank == maxrank) EXIT
        DO i = 1, nref
            part_cor = data(p, i)
            CALL linfit(part_cor,x,ntest,ssq_s(i))
        ENDDO
        ssq = sum(ssq_s(:))
        IF(ssq < ssq_g) THEN
            ssq_g = ssq
            test_order = p
            ssq_best = ssq_s
        ENDIF
    ENDDO

END SUBROUTINE perm_tj

subroutine perm_tj_successor ( n, p, rank )

!*****************************************************************************80
!
!! PERM_TJ_SUCCESSOR computes the Trotter-Johnson permutation successor.
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    27 July 2011
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Donald Kreher, Douglas Simpson,
!    Combinatorial Algorithms,
!    CRC Press, 1998,
!    ISBN: 0-8493-3988-X,
!    LC: QA164.K73.
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) N, the number of values being permuted.
!    N must be positive.
!
!    Input/output, integer ( kind = 4 ) P(N), describes the permutation.
!    P(I) is the item which is permuted into the I-th place
!    by the permutation.
!
!    Input/output, integer ( kind = 4 ) RANK, the rank.
!    If RANK = -1 on input, then the routine understands that this is
!    the first call, and that the user wishes the routine to supply
!    the first element in the ordering, which has RANK = 0.
!    In general, the input value of RANK is increased by 1 for output,
!    unless the very last element of the ordering was input, in which
!    case the output value of RANK is 0.
!
  implicit none

  integer ( kind = 4 ) n

  integer ( kind = 4 ) d
  logical done
  integer ( kind = 4 ) i
  integer ( kind = 4 ) m
  integer ( kind = 4 ) p(n)
  integer ( kind = 4 ) par
  integer ( kind = 4 ) q(n)
  integer ( kind = 4 ) rank
  integer ( kind = 4 ) st
  integer ( kind = 4 ) temp
!
!  Return the first element.
!
  if ( rank == -1 ) then
    call i4vec_indicator ( n, p )
    rank = 0
    return
  end if

  st = 0
  q(1:n) = p(1:n)
  done = .false.
  m = n

  do while ( 1 < m .and. .not. done )

    d = 1
    do while ( q(d) /= m )
      d = d + 1
    end do

    do i = d, m - 1
      q(i) = q(i+1)
    end do

    call perm_parity ( m - 1, q, par )

    if ( par == 1 ) then

      if ( d == m ) then
        m = m - 1
      else
        temp      = p(st+d)
        p(st+d)   = p(st+d+1)
        p(st+d+1) = temp
        done = .true.
      end if

    else

      if ( d == 1 ) then
        m = m - 1
        st = st + 1
      else
        temp      = p(st+d)
        p(st+d)   = p(st+d-1)
        p(st+d-1) = temp
        done = .true.
      end if

    end if

  end do
!
!  Last element was input.  Return first one.
!
  if ( m == 1 ) then
    call i4vec_indicator ( n, p )
    rank = 0
    return
  end if

  rank = rank + 1

  return
end

subroutine perm_tj_unrank ( rank, n, p )

!*****************************************************************************80
!
!! PERM_TJ_UNRANK computes the permutation of given Trotter-Johnson rank.
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    17 January 1999
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Donald Kreher, Douglas Simpson,
!    Combinatorial Algorithms,
!    CRC Press, 1998,
!    ISBN: 0-8493-3988-X,
!    LC: QA164.K73.
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) RANK, the rank of the permutation.
!
!    Input, integer ( kind = 4 ) N, the number of values being permuted.
!    N must be positive.
!
!    Output, integer ( kind = 4 ) P(N), describes the permutation.
!
  implicit none

  integer ( kind = 4 ) n

  integer ( kind = 4 ) i
  integer ( kind = 4 ) i4_factorial
  integer ( kind = 4 ) j
  integer ( kind = 4 ) k
  integer ( kind = 4 ) jhi
  integer ( kind = 4 ) nperm
  integer ( kind = 4 ) p(n)
  integer ( kind = 4 ) rank
  integer ( kind = 4 ) r1
  integer ( kind = 4 ) r2

  INTEGER (KIND = 8) i8
!
!  Check.
!
  if ( n < 1 ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'PERM_TJ_UNRANK - Fatal error!'
    write ( *, '(a)' ) '  Input N is illegal.'
    stop
  end if

  call perm_enum ( n, nperm )

  if ( rank < 0 .or. nperm < rank ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'PERM_TJ_UNRANK - Fatal error!'
    write ( *, '(a)' ) '  The input rank is illegal.'
    stop
  end if

  p(1) = 1
  r2 = 0														! rank = 3 & n = 3 --> p = 2, 1, 0

  do j = 2, n													!j = 2					!j = 3
!
!  Replace this ratio of factorials!
!
    IF(rank == 592 .and. r2 == 59 .and. jhi == 4 .and. k == 5 .and. j == 10 .and. i == 3) THEN
        i = k
    ENDIF
    i8 = INT(rank, KIND=8) * INT(i4_factorial ( j ), KIND=8)
    r1 = i8/i4_factorial ( n )
    k = r1 - j * r2

    if ( mod ( r2, 2 ) == 0 ) then
      jhi = j - k
    else
      jhi = k + 1
    end if

    do i = j - 1, jhi, -1
      p(i+1) = p(i)
    end do
    p(jhi) = j

    r2 = r1

  end do

  return
end

subroutine perm_parity ( n, p, parity )

!*****************************************************************************80
!
!! PERM_PARITY computes the parity of a permutation.
!
!  Discussion:
!
!    The routine requires the use of a temporary array.
!
!    A permutation is called "even" or "odd", depending on whether
!    it is equivalent to an even or odd number of pairwise
!    transpositions.  This is known as the "parity" of the
!    permutation.
!
!    The "sign" of a permutation is +1 if it has even parity,
!    and -1 if it has odd parity.
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    27 July 2011
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Donald Kreher, Douglas Simpson,
!    Combinatorial Algorithms,
!    CRC Press, 1998,
!    ISBN: 0-8493-3988-X,
!    LC: QA164.K73.
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) N, the number of values being permuted.
!    N must be positive.
!
!    Input, integer ( kind = 4 ) P(N), describes the permutation.
!    P(I) is the item which is permuted into the I-th place
!    by the permutation.
!
!    Output, integer ( kind = 4 ) PARITY, the parity of the permutation.
!    0, the permutation has even parity.
!    1, the permutation has odd parity.
!
  implicit none

  integer ( kind = 4 ) n

  integer ( kind = 4 ) a(n)
  integer ( kind = 4 ) c
  integer ( kind = 4 ) i
  integer ( kind = 4 ) j
  integer ( kind = 4 ) parity
  integer ( kind = 4 ) p(n)

  a(1:n) = 0

  c = 0

  do j = 1, n

    if ( a(j) == 0 ) then

      c = c + 1
      a(j) = 1
      i = j

      do while ( p(i) /= j )
        i = p(i)
        a(i) = 1
      end do

    end if

  end do

  parity = mod ( n - c, 2 )

  return
end

subroutine perm_enum ( n, nperm )

!*****************************************************************************80
!
!! PERM_ENUM enumerates the permutations on N digits.
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    23 January 1999
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) N, the number of values being permuted.
!    N must be nonnegative.
!
!    Output, integer ( kind = 4 ) NPERM, the number of distinct elements.
!
  implicit none

  integer ( kind = 4 ) i4_factorial
  integer ( kind = 4 ) n
  integer ( kind = 4 ) nperm

  nperm = i4_factorial ( n )

  return
end

function i4_factorial ( n )

!*****************************************************************************80
!
!! I4_FACTORIAL computes the factorial of an I4.
!
!  Formula:
!
!    Factorial ( N ) = Product ( 1 <= I <= N ) I
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    16 January 1999
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) N, the argument of the factorial function.
!    If N is less than 1, I4_FACTORIAL is returned as 1.
!
!    Output, integer ( kind = 4 ) I4_FACTORIAL, the factorial of N.
!
  implicit none

  integer ( kind = 4 ) i
  integer ( kind = 4 ) i4_factorial
  integer ( kind = 4 ) n

  i4_factorial = 1

  do i = 1, n
    i4_factorial = i4_factorial * i
  end do

  return
end
