SUBROUTINE perm_tj_reporting(resultfile, data, nref, ntest)
    !#############################################################################
    ! Perm_tj_reporting returns the minimal global SSQ for all test set permutations,
    ! the optimal order of the test factors, and for each reference
    ! factor the SSQ for the optimal rank order.
    !
    ! Uses the Trotter-Johnson permutation algorithm.
    !
    ! author: Florian Boecker   date: 05/01/2012
    ! http://complex-systems.uni-muenster.de/
    ! Licensing:  This code is distributed under the GNU GPL license.
    !
    ! parameters:
    !   name        i/o     datatype      description
    !   ----------------------------------------------------------------------------
    !   resultfile  input   character(:)  - the complete path of the file to be
    !                                          write out to/ will be modified here
    !   data        input   real(:,:)     - the correlation matrix
    !   nref        input   integer       - number of factors in the reference set
    !   ntest       input   integer       - number of factors in the test set
    !################################################################################

    !prevent implicite variable declaration
    IMPLICIT NONE

    !declaration
    !   parameters
    REAL, DIMENSION(ntest, nref) :: data !the correlation matrix (input)
    INTEGER nref        !number of factors in the reference set (input)
    INTEGER ntest       !number of factors in the test set (input)
    REAL, DIMENSION(nref) :: ssq_best    !vector with the ssqs for every reference factor resulting in the optimal ssq (output)
    REAL ssq_g       !sum of ssq_best: the lowest and ergo optimal ssq (output)
    INTEGER, DIMENSION(ntest) :: test_order !the order of test factors that resulted in the optimal ssq (output)
    !   varibales for internal use
    INTEGER rank    !rank of the permuation
    INTEGER, DIMENSION(ntest) :: p  !permutation
    REAL, DIMENSION(nref) :: ssq_s !vector with the ssq for every reference factor
    REAL ssq    ! the sum of squares of a permutation
    INTEGER i, j   !loop control
    REAL, DIMENSION(ntest) ::  part_cor !set of correlation values picked from the correlation matrix according to the permutation
    INTEGER (kind = 8) maxrank !rank that terminates the permutation
    REAL, DIMENSION(ntest) ::  x
    REAL, DIMENSION(nref) :: a, a_g ! interceptions for all permutation and for the best
    REAL, DIMENSION(nref) :: b, b_g ! slopes for all permutation and for the best
    INTEGER iunit
    CHARACTER (len=200) :: record, resultfile, resultfilecor, resultfiletxt


    !initialisation
    resultfilecor = TRIM(resultfile) // '.cor'      ! resultfile will be extended by '.cor' or '.txt'
    resultfiletxt = TRIM(resultfile) // '.txt'

    rank = -1   !tells perm_tj_successor to create the first permuation (rank = 0) (1,2,3,...,ntest)
    ssq_g = huge(1)
    maxrank = 1
    DO i = 1, ntest
        maxrank = maxrank * i
        x(i) = real(i)
    END DO
    maxrank = maxrank/2

    !calculation
    DO
        CALL perm_tj_successor(ntest, p, rank)
        IF(rank == maxrank) EXIT
        DO i = 1, nref
            part_cor = data(p, i)
            CALL linfit_reporting(part_cor,x,ntest,ssq_s(i),a(i),b(i))
        ENDDO
        ssq = sum(ssq_s(:))
        IF(ssq < ssq_g) THEN
            ssq_g = ssq
            test_order = p
            ssq_best = ssq_s
            a_g = a
            b_g = b
        ENDIF
    ENDDO

    !report cor matrix
    CALL get_unit(iunit)
    OPEN (unit = iunit, file = resultfilecor)
    DO i = 1, ntest
        record = ' '
        DO j = 1, nref
            CALL csv_record_append_re(data(i,j), record)
        ENDDO
        WRITE(iunit,'(a)') trim(record)
    ENDDO
    CLOSE(iunit)

    !report a and b of y=ax+b      for each regression graph a pair of values  = number of reference situations
    CALL get_unit(iunit)
    OPEN (unit = iunit, file = resultfiletxt)
    record = 'a:'
    DO i = 1, nref
        CALL csv_record_append_re(a_g(i), record)
    ENDDO
    WRITE(iunit,'(a)') trim(record)
    record = 'b:'
    DO i = 1, nref
        CALL csv_record_append_re(b_g(i), record)
    ENDDO
    WRITE(iunit,'(a)') trim(record)
    CLOSE(iunit)

END SUBROUTINE perm_tj_reporting

SUBROUTINE linfit_reporting(x,y,nr,ssq,a,b)
    ! ###############################################################
    ! Performes a linear fit based on a least square astimation,
    ! calculates the sum of squared y-residuals and reports back
    ! the intercept and the slope
    !
    !
    ! author: eberhard korsching   date: adapted on 26/04/2012
    !
    ! parameters:
    !   name        i/o     datatype      description
    !   ----------------------------------------------------------------------------
    !   x           input   real(:)       - vector with the x-values
    !   y           input   real(:)       - vector with the y-values
    !   nr          input   integer       - number of values in x and y
    !   ssq         output  real          - sum of squared y-residuals
    !   a           output  real          - intercept
    !   b           output  integer(:)    - slope
    ! ###############################################################

    !prevent implicite variable declaration
    IMPLICIT NONE

    ! linear fit based on least square estimation - ssq in y ! - important for distribution
    INTEGER nr
    REAL x(nr), y(nr), nra, ei(nr), xbar, ybar, b, a, ssq
    nra = real(nr)
    xbar = sum(x)/nra !average
    ybar = sum(y)/nra !average
    ! estimator for b
    b = sum((x-xbar)*(y-ybar))/sum((x-xbar)**2)
    ! estimator for a
    a = ybar-b*xbar
    ! estimator for the theoretical residuals
    ei = y-a-b*x
    ! ei : distances from the fittingcourve to the original value - von der Fitgeraden aus gesehen bzgl. Vorzeichen (- nach unten, + nach oben)
    ssq = sum(ei**2)
END SUBROUTINE linfit_reporting
