PROGRAM perm2011f90_comb_mpi
    !################################################################################
    ! activity:
    ! - reads in a csv file
    ! - devides the file collumnwise in reference and test factors
    ! - opens combinatorial space of reference sets
    ! - opens corresponding combinatorial space of test sets
    ! - devides the following work between the available processes
    !   * calculates a correlation matrix for every reference to test set
    !   * permutates the orders of test factors for every correlation matrix
    !   * creates a linear regression for every test factor permutation
    !   * calculates the ssq (sum of squared y deviations of the values to the
    !     regression line) for every linear regression
    ! - writes the results for every reference to test set combination to file
    !   (results are stored tab separated)
    !   * best (lowest) ssq
    !   * reference set
    !   * test set
    !   * order of test factors that produced the best ssq
    !
    ! arguments:
    ! #     obligation      characterization
    !--------------------------------------------------------------------------------
    ! 1.    mandatory       - path and name of the tab separated file for input
    ! 2.    mandatory       - path and name of the mapping file
    ! 3.    mandatory       - path and name of the file the result should be writen to
    ! 4.    mandatory       - path and name of the file that should be used for logging
    !                         information about the job
    ! 5.    optional        - number of combinations (if the number is not specified
    !                         or higher than the maximum number or of possible
    !                         combinations within the reference set (only) all possible
    !                         combinations are calculated
    ! 6.    optional        - number of reference factors
    !                         default = 6
    ! 7.    optional        - number factors differing from the original combination
    !                         default = number of reference factors
    !
    ! author: Florian Boecker   date: 16/07/2012, concept,idea: Eberhard Korsching
    ! http://complex-systems.uni-muenster.de/
    ! Licensing:  This code is distributed under the GNU GPL license.
    !################################################################################

    !MPI
    use mpi

    !prevent implicite variable declaration
    IMPLICIT NONE

    !Declaration ####################################################################

    !   parallelism related variables
    INTEGER error   !Mpi error codes
    INTEGER id  !ProcessID
    INTEGER p   !total number of processes
    INTEGER status(MPI_STATUS_SIZE)     !MPI status codes
    INTEGER tag !mpi message id, used to connect send and recieve commands, has to be equal in both commands
    INTEGER, ALLOCATABLE, DIMENSION(:,:) :: task_distribution   !first and last task for every process

    !   input realated variables
    INTEGER iargc ! Intrinsic function IARGS to retrive the number of arguments the programm was called with
    INTEGER numarg ! number of arguments the programm was called with
    CHARACTER (len = 10) argk ! for the argument retrival of the number of markers in the reference set
    LOGICAL :: file_exists ! recives the result of the intrinsic function INQUIRE to check the existance of the data file
    CHARACTER (len=200) :: infile, mapfile, resultfile, logfile     ! path and name of the in, result and log file
    INTEGER nrow      ! number of lines in the datafile
    INTEGER ncol      ! number of columns in the datafile
    INTEGER nrecord   ! number of data records
    INTEGER nfactor   ! number of factors in the datafile
    CHARACTER (len=10), ALLOCATABLE :: header(:)  ! contains the names of the Markers
    REAL, ALLOCATABLE, DIMENSION(:,:) :: rawdata   ! the raw data from the data file
    REAL, ALLOCATABLE, DIMENSION(:,:) :: data   ! the data from the data file w/o the first collumn
    INTEGER ctestcolumn ! loop control variable for the generation of the combinatorial space
    LOGICAL isinref     ! for the generation of the combinatorial space
    INTEGER maxc    !maximum number of combinations to be computated
    INTEGER, DIMENSION(100) :: mapping !mapping of the tma-data-file
    INTEGER, ALLOCATABLE, DIMENSION(:) :: seed  ! seed for random number generator
    INTEGER Seedlength  ! length of the seed for random number generator


    !   variables of global importance
    INTEGER n          ! total number of makers and length of the mapping
    INTEGER k          ! number of markers to be in the reference set
    INTEGER nsample    ! number of resampling procedures
    INTEGER c          ! number of test to ref combinations
    INTEGER z   !the number of things that can be different in each
    INTEGER (kind = 4), ALLOCATABLE, DIMENSION(:,:) :: cref    ! combinations of reference sets
    INTEGER (kind = 4), ALLOCATABLE, DIMENSION(:,:) :: ctest    ! combinations of test sets

    !   loop-controll variables
    INTEGER i, j, l

    !   calculated
    REAL, ALLOCATABLE, DIMENSION(:,:) :: corrmatrix    ! correlation matrix
    REAL ssq_g    ! sum of the ssq.best; g=global
    REAL, ALLOCATABLE, DIMENSION(:) :: ssq_best  ! minimum sum of squares of all reference factors for the best test order
    REAL, ALLOCATABLE, DIMENSION(:,:) :: ssqdata      ! all ssqs
    INTEGER, ALLOCATABLE, DIMENSION(:) :: test_order    ! optimal order for one combination
    INTEGER, ALLOCATABLE, DIMENSION(:,:) :: orderdata  ! all optimal orders

    !   output related variables
    CHARACTER (len=100) :: fmt  ! format variable to write the output
    CHARACTER (len = 300 ) record !Buffer string for output
    INTEGER logfileunit, resultfileunit     !identification ids for the output file units
    CHARACTER (len=8) :: date   !date retreved from intrinsic subroutine DATE_AND_TIME(date, time, zone, values)
    CHARACTER (len=10) :: time  !time retreved from intrinsic subroutine DATE_AND_TIME(date, time, zone, values)
    INTEGER, DIMENSION(8) :: timevector !int vector retreved from intrinsic subroutine DATE_AND_TIME(date, time, zone, values)
    CHARACTER (len=20) :: timechars(4)  !buffer vector for output of elapsed time
    DOUBLE PRECISION starttime, finishtime, elapsedtime !time stamps in seconds (year and month not included)
    LOGICAL univecfound ! triggers appropriate logging in case uniform marker vector

    !Initialisation ####################################################################

    !   Initialize MPI
    CALL MPI_Init(error)
    !   Get the number of processes
    CALL MPI_Comm_size(MPI_COMM_WORLD, p, error)
    !   Get the individual process ID
    CALL MPI_Comm_rank(MPI_COMM_WORLD, id, error)

    !   argument retrival
    numarg = iargc( )
    IF(numarg < 4) THEN
        IF(id == 0) THEN
            WRITE(*,*) 'not enough arguments submitted'
            WRITE(*,*) 'scheme: infile mapfile resultfile logfile (maxc) (nref)'
            WRITE(*,*) 'the specification of the files is mandatory'
            WRITE(*,*) 'programm cancled'
        ENDIF
        STOP
    ELSE
        CALL getarg (1, infile)
        INQUIRE(FILE=infile, EXIST=file_exists)
        IF(.NOT.file_exists) THEN
            IF(id == 0) THEN
                WRITE (*,*) 'input file not found'
                WRITE (*,*) 'programm cancled'
            ENDIF
            STOP
        ENDIF
        CALL getarg (2, mapfile)
        INQUIRE(FILE=mapfile, EXIST=file_exists)
        IF(.NOT.file_exists) THEN
            IF(id == 0) THEN
                WRITE (*,*) 'mapping file not found'
                WRITE (*,*) 'programm cancled'
            ENDIF
            STOP
        ENDIF
        CALL getarg (3, resultfile)
        CALL getarg (4, logfile)
        IF(numarg > 4) THEN
            CALL getarg (5, argk)
            READ(argk,*) maxc
        ELSE
            IF(id == 0) THEN
                WRITE (*,*) 'No maximum number of combinations specified'
                WRITE (*,*) 'All combinations are computated'
            ENDIF
            maxc = huge(1)
        ENDIF
        IF(numarg > 5) THEN
            CALL getarg (6, argk)
            READ(argk,*) k
        ELSE
            IF(id == 0) THEN
                WRITE (*,*) 'No reference set size specified'
                WRITE (*,*) 'Default value of 6 is used'
            ENDIF
            k = 6 ! default number of markers to be in the reference set
        ENDIF
        IF(numarg > 6) THEN
            CALL getarg (7, argk)
            READ(argk,*) z
            IF(z < 1 .OR. z > k) THEN
                IF(id == 0) THEN
                    WRITE (*,*) 'The specified value for z (maximum number of factors by which'
                    WRITE (*,*) 'the combinations are allowed to differ from the original'
                    WRITE (*,*) 'combination) is illegal.'
                    WRITE (*,*) 'Number of refernce factors is used.'
                ENDIF
                z = k
            ENDIF
        ELSE
            IF(id == 0) THEN
                WRITE (*,*) 'Number of factors that can differ from the original combination not specified'
                WRITE (*,*) 'Number of refernce factors is used'
            ENDIF
            z = k
        ENDIF
        IF(numarg > 7 .AND. id == 0) THEN
            WRITE (*,*) 'Arguments after size of reference set irgnored'
        ENDIF
    ENDIF

    !open logfile (only in master process)
    IF(id == 0) THEN
        CALL get_unit(logfileunit)
        OPEN (unit = logfileunit, file = logfile)
        CALL DATE_AND_TIME(date, time, values = timevector)
        starttime = timevector(3)*24*3600+timevector(5)*3600+timevector(6)*60+timevector(7)+DBLE(timevector(8))/DBLE(1000)
        WRITE(logfileunit,*) 'Starting job: ',date(7:8),'.',date(5:6),'.',date(1:4),' ',time(1:2),':',time(3:4),':',time(5:10)
    ENDIF

    !Map the data file
    CALL map_tma_data_file(infile, mapfile, ncol, nrow, nfactor, nrecord, mapping, n)

    !Calculate the number test to ref combinations (-> number of tasks)
    CALL comb_part_enum(n, k, 1, z, c)
    IF(maxc < c) c = maxc !limit the number of combinations if dictated so by the argument

    !Determine the distribution of tasks
    ALLOCATE (task_distribution(p,2))
    CALL task_division(c, p, task_distribution)

    !Allocation
    ALLOCATE (header(n)) ! allocate space for the names of the Markers
    ALLOCATE (rawdata(nrecord, n))  ! allocate space for the raw data from the data file
    ALLOCATE (cref(c, k))      ! allocate space for the combinations of reference sets
    ALLOCATE (ctest(c, n-k))      ! allocate space for the combinations of test sets
    ALLOCATE (ssq_best(k)) ! allocate space for ssq_best
    ALLOCATE (test_order(n-k))
    ALLOCATE (corrmatrix(n-k, k))  ! allocate space for the correlation matrix
    ALLOCATE (ssqdata(c, k+1))
    ALLOCATE (orderdata(c, n-k))

    !Read in the datafile
    CALL read_tma_data_file(infile, ncol, nrow, nfactor, nrecord, mapping, n, header, rawdata)

    IF(z > n-k) THEN
        WRITE (*,*) 'The specified value for z (maximum number of factors by which'
        WRITE (*,*) 'the combinations are allowed to differ from the original'
        WRITE (*,*) 'combination) is higher than the size of the test set.'
        WRITE (*,*) 'Size of test set is used.'
        z = n-k
    ENDIF

    !Open combinatorial space of reference sets
    CALL comb_part(n, k, 1, z, cref, c)

    !Open combinatorial space of test sets corresponding to the reference sets
    DO i = 1, c               ! c rows in cref und ctest
        ctestcolumn = 1       ! n-k columns in ctest
        DO j = 1, n           ! n different markers
            isinref = .false.
            DO l = 1, k       ! k columns in cref
                IF(j == cref(i,l)) THEN
                    isinref = .true.
                ENDIF
            END DO
            IF(.not. isinref) THEN
                ctest(i,ctestcolumn) = j
                ctestcolumn = ctestcolumn + 1
            ENDIF
        END DO
    END DO

    !Initialisation of the seed for random number generation based on the current date and time
    !Important for the calculation of the pearson correlation in the case of a uniform vector
    CALL RANDOM_SEED(SIZE=seedlength)
    ALLOCATE(seed(seedlength))
    CALL get_seed(seed(1))
    seed(:) =  seed(1)
    CALL RANDOM_SEED (PUT = seed(:))

    !logging stuff (only in master process)
    IF(id == 0) THEN
        WRITE(logfileunit,*)
        WRITE(logfileunit,'(A13, TR2, A75)') 'file:', infile
        WRITE(logfileunit,'(A13, TR2, A75)') 'mapfile:', mapfile
        WRITE(logfileunit,'(A13, TR2, I6)') 'rows:', nrow
        WRITE(logfileunit,'(A13, TR2, I6)') 'columns:', ncol
        WRITE(logfileunit,'(A13, TR2, I6)') 'factors:', nfactor
        WRITE(logfileunit,'(A13, TR2, I6)') 'n:', n
        WRITE(logfileunit,'(A13, TR2, I6)') 'reference:', k
        WRITE(logfileunit,'(A13, TR2, I6)') 'z:', z
        WRITE(logfileunit,'(A13, TR2, I6)') 'test:', n-k
        WRITE(logfileunit,'(A13, TR2, I6)') 'combinations:', c
        WRITE(fmt,*) '(A13, TR2,', n, 'A)'
        WRITE(logfileunit,fmt) 'marker:', header(:)
        header(:) = ' '
        univecfound = .FALSE.
        DO i = 1, n
            IF(SUM(rawdata(:,i)-SUM(rawdata(:,i))/nrecord) == 0) THEN
                header(i) = 'X'
                univecfound = .TRUE.
            ENDIF
        ENDDO
        IF(univecfound == .TRUE.) THEN
            WRITE(logfileunit,fmt) 'uniform:', header(:)
            WRITE(logfileunit,*) 'Uniform marker vector(s) found!'
            WRITE(logfileunit,*) 'The Person correlation can not be calculated correctly with this marker(s).'
            WRITE(logfileunit,*) 'Instead a correlation value near to 0 is randomly generated.'
        ENDIF
        WRITE(logfileunit,*)
        WRITE(logfileunit,'(A13, TR2, I6)') 'processes:', p
        WRITE(logfileunit,*)
        WRITE(logfileunit,*) 'distribution of tasks:'
        WRITE(logfileunit,*) 'processes     first task      last task'
        DO i = 1, p
            WRITE(logfileunit,'(3X I4, 8X I6, 11X I6)') i, task_distribution(i,1), task_distribution(i,2)
        ENDDO
        WRITE(logfileunit,*)
        CALL DATE_AND_TIME(date, time)
        WRITE(logfileunit,*) 'Starting calculations: ',date(7:8),'.',date(5:6),'.',date(1:4),' ',time(1:2),':',time(3:4),':',time(5:10)
    ENDIF

    ! Calculation #################################################################
    DO i = task_distribution(id+1,1), task_distribution(id+1,2)
        ! Calculate the matrix of correlation-coeffientients
        CALL computedistance(rawdata, cref(i,:), ctest(i,:), n, k, nrecord, corrmatrix)
        ! Calculate the best ssq for each reference factor, the minimal global ssq an the optimal order of the test factors
        CALL perm_tj(corrmatrix, k, n-k, ssq_best, ssq_g, test_order)
        ssqdata(i,1:k) = ssq_best(:)
        ssqdata(i,k+1) = ssq_g
        orderdata(i,:) = test_order(:)
    END DO

    !Log end of calculation (only in master process)
    IF(id == 0) THEN
        CALL DATE_AND_TIME(date, time)
        WRITE(logfileunit,*) 'Finished calculation and start collecting data: ',date(7:8),'.',date(5:6),'.',date(1:4),' ',time(1:2),':',time(3:4),':',time(5:10)
    ENDIF


    !Collect Data
    tag = 0
    IF(id == 0) THEN
        DO i = 2, p
            DO j = task_distribution(i,1), task_distribution(i,2)
                !recive
                CALL MPI_Recv(ssqdata(j,:), k+1, MPI_INTEGER, i-1, tag, MPI_COMM_WORLD, status, error)
                tag = tag + 1
                CALL MPI_Recv(orderdata(j,:), n-k, MPI_INTEGER, i-1, tag, MPI_COMM_WORLD, status, error)
                tag = tag + 1
            END DO
        END DO
    ELSE
        DO i = 2, p
            DO j = task_distribution(i,1), task_distribution(i,2)
                !send
                IF(id == i-1) THEN
                    CALL MPI_Send(ssqdata(j,:), k+1, MPI_INTEGER, 0, tag , MPI_COMM_WORLD, error)
                ENDIF
                tag = tag + 1
                IF(id == i-1) THEN
                    CALL MPI_Send(orderdata(j,:), n-k, MPI_INTEGER, 0, tag , MPI_COMM_WORLD, error)
                ENDIF
                tag = tag + 1
            END DO
        ENDDO
    ENDIF

    !close mpi
    call MPI_Finalize(error)

    ! Output (only executed in master process)
    IF(id == 0) THEN
        CALL get_unit(resultfileunit)
        OPEN (unit = resultfileunit, file = resultfile)
        DO i = 1, c
            record = ' '
            CALL csv_record_append_re (ssqdata(i,k+1), record) !ssqg
            DO j = 1, n
                IF(j <= k) THEN
                    CALL csv_record_append_i4 (cref(i,j), record)   !ref set
                ELSE
                    CALL csv_record_append_i4 (ctest(i,j-k), record)    !test set
                ENDIF
            ENDDO
            DO j = 1, n-k
                CALL  csv_record_append_i4 (orderdata(i,j), record) !optimal order
            ENDDO
            WRITE(resultfileunit,'(a)') trim(record)
        END DO
    ENDIF

    !Log elapsed time (only in master process)
    IF(id == 0) THEN
        CALL DATE_AND_TIME(date, time, values = timevector)
        WRITE(logfileunit,*) 'Finished collecting data: ',date(7:8),'.',date(5:6),'.',date(1:4),' ',time(1:2),':',time(3:4),':',time(5:10)
        ! calculate elapsed time
        finishtime = timevector(3)*24*3600+timevector(5)*3600+timevector(6)*60+timevector(7)+DBLE(timevector(8))/DBLE(1000)
        elapsedtime = finishtime-starttime
        ! days
        WRITE (timechars(1),*) INT(elapsedtime/(24*3600))
        timechars(1) = adjustl(timechars(1))
        ! hours
        WRITE (timechars(2),*) INT(mod(elapsedtime,DBLE(24*3600))/3600)
        timechars(2) = adjustl(timechars(2))
        ! minutes
        WRITE (timechars(3),*) INT(mod(mod(elapsedtime,DBLE(24*3600)),DBLE(3600))/60)
        timechars(3) = adjustl(timechars(3))
        ! seconds
        WRITE (timechars(4),'(F6.3)') mod(mod(mod(elapsedtime,DBLE(24*3600)),DBLE(3600)),DBLE(60))
        timechars(4) = adjustl(timechars(4))
        ! write elapsed time to file
        WRITE(logfileunit,*)
        WRITE(logfileunit,*) 'Finished computation after '//trim(timechars(1))//' days, '//trim(timechars(2))//' hours, '//trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
        !close log file
        CLOSE(logfileunit)
        !report to console for performance tests
        WRITE(record,*) p
        record = ADJUSTL(record)
        IF (timechars(1) /= '0') THEN
            IF(p == 1) THEN
                WRITE(*,*) 'Finished job on '//trim(record)//' process after '//trim(timechars(1))//' days, '//trim(timechars(2))//' hours, '//trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
            ELSE
                WRITE(*,*) 'Finished job on '//trim(record)//' processes after '//trim(timechars(1))//' days, '//trim(timechars(2))//' hours, '//trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
            ENDIF
        ELSE
            IF (timechars(2) /= '0') THEN
                IF(p == 1) THEN
                    WRITE(*,*) 'Finished job on '//trim(record)//' process after '//trim(timechars(2))//' hours, '//trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
                ELSE
                    WRITE(*,*) 'Finished job on '//trim(record)//' processes after '//trim(timechars(2))//' hours, '//trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
                ENDIF
            ELSE
                IF (timechars(3) /= '0') THEN
                    IF(p == 1) THEN
                        WRITE(*,*) 'Finished job on '//trim(record)//' process after '//trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
                    ELSE
                        WRITE(*,*) 'Finished job on '//trim(record)//' processes after '//trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
                    ENDIF
                ELSE
                    IF(p == 1) THEN
                        WRITE(*,*) 'Finished job on '//trim(record)//' process after '//trim(timechars(4))//' seconds'
                    ELSE
                        WRITE(*,*) 'Finished job on '//trim(record)//' processes after '//trim(timechars(4))//' seconds'
                    ENDIF
                ENDIF
            ENDIF
        ENDIF
    ENDIF

END PROGRAM perm2011f90_comb_mpi
