PROGRAM perm2011f90_comb_omp
    !################################################################################
    ! activity
    ! - reads in a csv file
    ! - devides the file collumnwise in reference and test factors
    ! - opens combinatorial space of reference sets
    ! - opens corresponding combinatorial space of test sets
    ! - devides the following work between the available processes
    !   * calculates a correlation matrix for every reference to test set
    !   * permutates the orders of test factors for every correlation matrix
    !   * creates a linear regression for every test factor permutation
    !   * calculates the ssq (sum of squared y deviations of the values to the
    !     regression line) for every linear regression
    ! - writes the results for every reference to test set combination to file
    !   (results are stored tab separated)
    !   * best (lowest) ssq
    !   * reference set
    !   * test set
    !   * order of test factors that produced the best ssq
    !
    ! arguments:
    ! #     obligation      characterization
    !--------------------------------------------------------------------------------
    ! 1.    mandatory       - path and name of the tab separated file for input
    ! 2.    mandatory       - path and name of the mapping file
    ! 3.    mandatory       - path and name of the file the result should be writen to
    ! 4.    mandatory       - path and name of the file that should be used for logging
    !                         information about the job
    ! 5.    optional        - number of combinations (if the number is not specified
    !                         or highert han the maximum number or of possible
    !                         combinations within the reference set (only) all possible
    !                         combinations are calculated
    ! 6.    optional        - number of reference factors
    !                         default = 6
    ! 7.    optional        - number factors differing from the original combination
    !                         default = number of reference factors
    ! 8.    optional        - number of threads openMP should use for parallel computation
    !                         default = number of all available processors
    !
    !
    ! author: Florian Boecker   date: 16/07/2012, concept,idea: Eberhard Korsching
    ! http://complex-systems.uni-muenster.de/
    ! Licensing:  This code is distributed under the GNU GPL license.
    !################################################################################

    !OpenMP
    use omp_lib

    !prevent implicite variable declaration
    IMPLICIT NONE

    !Decleration ################################################################################
    INTEGER n          ! total number of makers
    INTEGER k          ! number of markers to be in the reference set
    INTEGER i, j, l, ctestcolumn ! loop-controll variables
    INTEGER c          ! number of test to ref combinations
    INTEGER nrow       ! number of lines in the datafile
    INTEGER ncol       ! number of columns in the datafile
    INTEGER nfactor    ! number of factors in the datafile
    INTEGER nrecord    ! number of data records
    CHARACTER (len=200) :: infile, mapfile, resultfile, logfile     ! path and name of the in, map, result and log file
    CHARACTER (len=100) :: fmt  ! format variable to write the output
    CHARACTER (len=10), ALLOCATABLE :: header(:)  ! contains the names of the Markers
    REAL, ALLOCATABLE, DIMENSION(:,:) :: rawdata   ! the raw data from the data file
    INTEGER (kind = 4), ALLOCATABLE, DIMENSION(:,:) :: cref    ! combinations of reference sets
    INTEGER (kind = 4), ALLOCATABLE, DIMENSION(:,:) :: ctest    ! combinations of test sets
    LOGICAL isinref
    REAL, ALLOCATABLE, DIMENSION(:) :: ssq_best  ! minimum sum of squares of all reference factors for the best test order
    REAL ssq_g    ! sum of the ssq.best; g=global
    INTEGER, ALLOCATABLE, DIMENSION(:) :: test_order
    CHARACTER (len=8) :: date
    CHARACTER (len=10) :: time
    INTEGER, DIMENSION(8) :: timevector
    CHARACTER (len=20) :: timechars(4)
    DOUBLE PRECISION starttime, finishtime, elapsedtime
    INTEGER iargc ! Intrinsic function IARGS to retrive the number of arguments the programm was called with
    INTEGER numarg ! number of arguments the programm was called with
    CHARACTER (len = 10) argk ! for the argument retrival of the number of markers in the reference set
    LOGICAL :: file_exists ! recives the result of the intrinsic function INQUIRE to check the existance of the data file
    REAL, ALLOCATABLE, DIMENSION(:,:) :: corrmatrix    ! correlation matrix
    REAL, ALLOCATABLE, DIMENSION(:,:) :: ssqdata      ! all ssqs
    INTEGER, ALLOCATABLE, DIMENSION(:,:) :: orderdata  ! all optimal orders
    CHARACTER (len = 200 ) record !Buffer string for output
    INTEGER logfileunit, resultfileunit
    INTEGER maxc    !maximum number of combinations to be computated
    INTEGER, DIMENSION(100) :: mapping !mapping of the tma-data-file
    INTEGER, ALLOCATABLE, DIMENSION(:) :: seed  ! seed for random number generator
    INTEGER Seedlength  ! length of the seed for random number generator
    LOGICAL univecfound ! triggers appropriate logging in case uniform marker vector
    !parallelism related variables
    INTEGER max_threads ! number of threads openMP uses for parallel computation
    INTEGER proc_num    ! number of processors
    INTEGER z   !the number of things that can be different in each combination

    !   argument retrival
    numarg = iargc( )
    IF(numarg < 4) THEN
        WRITE(*,*) 'not enough arguments submitted'
        WRITE(*,*) 'scheme: infile mapfile resultfile logfile (maxc) (nref)'
        WRITE(*,*) 'the specification of the files is mandatory'
        WRITE(*,*) 'programm cancled'
        STOP
    ELSE
        CALL getarg (1, infile)
        INQUIRE(FILE=infile, EXIST=file_exists)
        IF(.NOT.file_exists) THEN
            WRITE (*,*) 'input file not found'
            WRITE (*,*) 'programm cancled'
            STOP
        ENDIF
        CALL getarg (2, mapfile)
        INQUIRE(FILE=mapfile, EXIST=file_exists)
        IF(.NOT.file_exists) THEN
            WRITE (*,*) 'mapping file not found'
            WRITE (*,*) 'programm cancled'
            STOP
        ENDIF
        CALL getarg (3, resultfile)
        CALL getarg (4, logfile)
        IF(numarg > 4) THEN
            CALL getarg (5, argk)
            READ(argk,*) maxc
        ELSE
            WRITE (*,*) 'No maximum number of combinations specified'
            WRITE (*,*) 'All combinations are computated'
            maxc = huge(1)
        ENDIF
        IF(numarg > 5) THEN
            CALL getarg (6, argk)
            READ(argk,*) k
        ELSE
            WRITE (*,*) 'No reference set size specified'
            WRITE (*,*) 'Default value of 6 is used'
            k = 6 ! default number of markers to be in the reference set
        ENDIF
        IF(numarg > 6) THEN
            CALL getarg (7, argk)
            READ(argk,*) z
            IF(z < 1 .OR. z > k) THEN
                WRITE (*,*) 'The specified value for z (maximum number of factors by which'
                WRITE (*,*) 'the combinations are allowed to differ from the original'
                WRITE (*,*) 'combination) is illegal.'
                WRITE (*,*) 'Number of refernce factors is used.'
                z = k
            ENDIF
        ELSE
            WRITE (*,*) 'Number of factors that can differ from the original combination not specified'
            WRITE (*,*) 'Number of refernce factors is used'
            z = k
        ENDIF
        IF(numarg > 7) THEN
            CALL getarg (8, argk)
            READ(argk,*) max_threads
            IF(max_threads > 0) THEN
                CALL omp_set_num_threads(max_threads) !set number of threads
            ELSE
                WRITE (*,*) 'Specified number of threads is illegal'
                WRITE (*,*) 'One thread per available processor is used'
                max_threads = omp_get_max_threads()
            ENDIF
        ELSE
            WRITE (*,*) 'No number of threads spezified'
            WRITE (*,*) 'One thread per available processor is used'
            max_threads = omp_get_max_threads()
        ENDIF
        IF(numarg > 8) THEN
            WRITE (*,*) 'Excess arguments irgnored'
        ENDIF
    ENDIF

    !open logfile
    CALL get_unit(logfileunit)
    OPEN (unit = logfileunit, file = logfile)

    CALL DATE_AND_TIME(date, time, values = timevector)
    starttime = timevector(3)*24*3600+timevector(5)*3600+timevector(6)*60+timevector(7)+DBLE(timevector(8))/DBLE(1000)
    WRITE(logfileunit,*) 'Starting job: ',date(7:8),'.',date(5:6),'.',date(1:4),' ',time(1:2),':',time(3:4),':',time(5:10)

    !Map the data file
    CALL map_tma_data_file(infile, mapfile, ncol, nrow, nfactor, nrecord, mapping, n)

    IF(z > n-k) THEN
        WRITE (*,*) 'The specified value for z (maximum number of factors by which'
        WRITE (*,*) 'the combinations are allowed to differ from the original'
        WRITE (*,*) 'combination) is higher than the size of the test set.'
        WRITE (*,*) 'Size of test set is used.'
        z = n-k
    ENDIF

    !Calculate the number test to ref combinations (-> number of tasks)
    CALL comb_part_enum(n, k, 1, z, c)
    IF(maxc < c) c = maxc !limt the number of combinations if dictated so by an argument

    !Determine the parameters regarding parallelism
    proc_num = omp_get_num_procs( )
    max_threads = omp_get_max_threads( )

    !Allocation
    ALLOCATE (header(n)) ! allocate space for the names of the Markers
    ALLOCATE (rawdata(nrecord, n))  ! allocate space for the raw data from the data file
    ALLOCATE (cref(c, k))      ! allocate space for the combinations of reference sets
    ALLOCATE (ctest(c, n-k))      ! allocate space for the combinations of test sets
    ALLOCATE (ssqdata(c, k+1))
    ALLOCATE (orderdata(c, n-k))

    !Read in the datafile
    CALL read_tma_data_file(infile, ncol, nrow, nfactor, nrecord, mapping, n, header, rawdata)

    !Open combinatorial space of reference sets
    CALL comb_part(n, k, 1, z, cref, c)

    !Open combinatorial space of test sets corresponding to the reference sets
    DO i = 1, c               ! c rows in cref und ctest
        ctestcolumn = 1       ! n-k columns in ctest
        DO j = 1, n           ! n different markers
            isinref = .false.
            DO l = 1, k       ! k columns in cref
                IF(j == cref(i,l)) THEN
                    isinref = .true.
                ENDIF
            END DO
            IF(.not. isinref) THEN
                ctest(i,ctestcolumn) = j
                ctestcolumn = ctestcolumn + 1
            ENDIF
        END DO
    END DO

    !Initialisation of the seed for random number generation based on the current date and time
    !Important for the calculation of the pearson correlation in the case of a uniform vector
    CALL RANDOM_SEED(SIZE=seedlength)
    ALLOCATE(seed(seedlength))
    CALL get_seed(seed(1))
    seed(:) =  seed(1)
    CALL RANDOM_SEED (PUT = seed(:))

    !logging stuff
    WRITE(logfileunit,*)
    WRITE(logfileunit,'(A13, TR2, A75)') 'file:', infile
    WRITE(logfileunit,'(A13, TR2, A75)') 'mapfile:', mapfile
    WRITE(logfileunit,'(A13, TR2, A75)') 'logfile:', logfile
    WRITE(logfileunit,'(A13, TR2, A75)') 'resultfile:', resultfile
    WRITE(logfileunit,'(A13, TR2, I6)') 'rows:', nrow
    WRITE(logfileunit,'(A13, TR2, I6)') 'columns:', ncol
    WRITE(logfileunit,'(A13, TR2, I6)') 'factors:', nfactor
    WRITE(logfileunit,'(A13, TR2, I6)') 'n:', n
    WRITE(logfileunit,'(A13, TR2, I6)') 'reference:', k
    WRITE(logfileunit,'(A13, TR2, I6)') 'z:', z
    WRITE(logfileunit,'(A13, TR2, I6)') 'test:', n-k
    WRITE(logfileunit,'(A13, TR2, I6)') 'combinations:', c
    WRITE(logfileunit,'(A13, TR2, I6)') 'processors:', proc_num
    WRITE(logfileunit,'(A13, TR2, I6)') 'threads:', max_threads
    WRITE(fmt,*) '(A13, TR2,', n, 'A)'
    WRITE(logfileunit,fmt) 'marker:', header(:)
    header(:) = ' '
    univecfound = .FALSE.
    DO i = 1, n
        IF(SUM(rawdata(:,i)-SUM(rawdata(:,i))/nrecord) == 0) THEN
            header(i) = 'X'
            univecfound = .TRUE.
        ENDIF
    ENDDO
    IF(univecfound == .TRUE.) THEN
        WRITE(logfileunit,fmt) 'uniform:', header(:)
        WRITE(logfileunit,*) 'Uniform marker vector(s) found!'
        WRITE(logfileunit,*) 'The Person correlation can not be calculated correctly with this marker(s).'
        WRITE(logfileunit,*) 'Instead a correlation value near to 0 is randomly generated.'
    ENDIF
    WRITE(logfileunit,*)
    CALL DATE_AND_TIME(date, time)
    WRITE(logfileunit,*) 'Starting calculations: ',date(7:8),'.',date(5:6),'.',date(1:4),' ',time(1:2),':',time(3:4),':',time(5:10)

    ! Calculation #################################################################
    !$omp parallel do private(ssq_g, corrmatrix, ssq_best, test_order)
    DO i = 1, c
        ALLOCATE (corrmatrix(n-k, k))  ! allocate space for the correlation matrix
        ALLOCATE (ssq_best(k)) ! allocate space for ssq_best
        ALLOCATE (test_order(n-k))
        ! Calculate the matrix of correlation-coeffientients
        CALL computedistance(rawdata, cref(i,:), ctest(i,:), n, k, nrecord, corrmatrix)
        ! Calculate the best ssq for each reference factor, the minimal global ssq an the optimal order of the test factors
        CALL perm_tj(corrmatrix, k, n-k, ssq_best, ssq_g, test_order)
        ssqdata(i,1:k) = ssq_best(:)
        ssqdata(i,k+1) = ssq_g
        orderdata(i,:) = test_order(:)
        DEALLOCATE(corrmatrix, ssq_best, test_order)
    END DO
    !$omp end parallel do

    !Log end of calculation
    CALL DATE_AND_TIME(date, time)
    WRITE(logfileunit,*) 'Finished calculation and start collecting data: ',date(7:8),'.',date(5:6),'.',date(1:4),' ',&
    time(1:2),':',time(3:4),':',time(5:10)

    !open resultfile
    CALL get_unit(resultfileunit)
    OPEN (unit = resultfileunit, file = resultfile)

    ! Output results
    DO i = 1, c
        record = ' '
        CALL csv_record_append_re (ssqdata(i,k+1), record) !ssqg
        DO j = 1, n
            IF(j <= k) THEN
                CALL csv_record_append_i4 (cref(i,j), record)   !ref set
            ELSE
                CALL csv_record_append_i4 (ctest(i,j-k), record)    !test set
            ENDIF
        ENDDO
        DO j = 1, n-k
            CALL  csv_record_append_i4 (orderdata(i,j), record) !optimal order
        ENDDO
        WRITE(resultfileunit,'(A)') trim(record)
    END DO

    !close result file
    CLOSE(resultfileunit)

    !Log elapsed time
    CALL DATE_AND_TIME(date, time, values = timevector)
    WRITE(logfileunit,*) 'Finished collecting data: ',date(7:8),'.',date(5:6),'.',date(1:4),' ',time(1:2),':'&
    ,time(3:4),':',time(5:10)
    ! calculate elapsed time
    finishtime = timevector(3)*24*3600+timevector(5)*3600+timevector(6)*60+timevector(7)+DBLE(timevector(8))/DBLE(1000)
    elapsedtime = finishtime-starttime
    ! days
    WRITE (timechars(1),*) INT(elapsedtime/(24*3600))
    timechars(1) = adjustl(timechars(1))
    ! hours
    WRITE (timechars(2),*) INT(mod(elapsedtime,DBLE(24*3600))/3600)
    timechars(2) = adjustl(timechars(2))
    ! minutes
    WRITE (timechars(3),*) INT(mod(mod(elapsedtime,DBLE(24*3600)),DBLE(3600))/60)
    timechars(3) = adjustl(timechars(3))
    ! seconds
    WRITE (timechars(4),'(F6.3)') mod(mod(mod(elapsedtime,DBLE(24*3600)),DBLE(3600)),DBLE(60))
    timechars(4) = adjustl(timechars(4))
    ! write elapsed time to file
    WRITE(logfileunit,*)
    WRITE(logfileunit,*) 'Finished computation after '//trim(timechars(1))//' days, '//trim(timechars(2))//' hours, '&
    //trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'

    !close log file
    CLOSE(logfileunit)

    !report to console for performance tests
    WRITE(record,*) max_threads
    record = ADJUSTL(record)
    IF (timechars(2) /= '0') THEN
            IF(max_threads == 1) THEN
                WRITE(*,*) 'Finished job on '//trim(record)//' thread after '//trim(timechars(2))//' hours, '&
                //trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
            ELSE
                WRITE(*,*) 'Finished job on '//trim(record)//' threads after '//trim(timechars(2))//' hours, '&
                //trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
            ENDIF
    ELSE
        IF (timechars(3) /= '0') THEN
            IF(max_threads == 1) THEN
                WRITE(*,*) 'Finished job on '//trim(record)//' thread after '//trim(timechars(3))//' minutes and '&
                //trim(timechars(4))//' seconds'
            ELSE
                WRITE(*,*) 'Finished job on '//trim(record)//' threads after '//trim(timechars(3))//' minutes and '&
                //trim(timechars(4))//' seconds'
            ENDIF
        ELSE
            IF(max_threads == 1) THEN
                WRITE(*,*) 'Finished job on '//trim(record)//' thread after '//trim(timechars(4))//' seconds'
            ELSE
                WRITE(*,*) 'Finished job on '//trim(record)//' threads after '//trim(timechars(4))//' seconds'
            ENDIF
        ENDIF
    ENDIF

END PROGRAM perm2011f90_comb_omp
