PROGRAM perm2011f90_resample_mpi
    !################################################################################
    ! activity
    ! - reads in a csv file
    ! - shuffles or bootstraps the data
    ! - devides the file collumnwise in reference and test factors
    ! - calculates a correlation matrix
    ! - permutates the orders of test factors for every resampling
    ! - calculates the ssq for every order permutation
    ! - writes the best (lowest) ssq and its order of test factors in a ouput file
    !
    ! arguments:
    ! #     obligation      characterization
    !--------------------------------------------------------------------------------
    ! 1.    mandatory       - path and name of the tab separated file for input
    ! 2.    mandatory       - path and name of the mapping file
    ! 3.    mandatory       - path and name of the file the result should be writen to
    ! 4.    mandatory       - path and name of the file that should be used for logging
    !                         information about the job
    ! 5.    mandatory       - number of shufflings
    ! 6.    optional        - resampling method
    !                         's' or 'S' for shuffling
    !                         'b' or 'B' for bootstrapping
    !                         default: shuffling
    ! 7.    optional        - number of reference factors
    !                         default = 6
    !
    ! author: Florian Boecker   date: 16/07/2012, concept,idea: Eberhard Korsching
    ! http://complex-systems.uni-muenster.de/
    ! Licensing:  This code is distributed under the GNU GPL license.
    !################################################################################


    !MPI
    use mpi

    !prevent implicite variable declaration
    IMPLICIT NONE

    !Declaration ####################################################################

    !   parallelism related variables
    INTEGER error   !Mpi error codes
    INTEGER id  !ProcessID
    INTEGER p   !total number of processes
    INTEGER status(MPI_STATUS_SIZE)     !MPI status codes
    INTEGER tag !mpi message id, used to connect send and recieve commands, has to be equal in both commands
    INTEGER, ALLOCATABLE, DIMENSION(:,:) :: task_distribution   !first and last task for every process
    INTEGER, ALLOCATABLE, DIMENSION(:) :: seed
    INTEGER Seedlength

    !   input realated variables
    INTEGER iargc ! Intrinsic function IARGS to retrive the number of arguments the programm was called with
    INTEGER numarg ! number of arguments the programm was called with
    CHARACTER (len = 10) argk ! for the argument retrival of the number of markers in the reference set
    LOGICAL :: file_exists ! recives the result of the intrinsic function INQUIRE to check the existance of the data file
    CHARACTER (len=1) :: resamplingmethod   !indicator for the method that is to be used for the resampling(s=shuffling, b=bootstrapping)
    CHARACTER (len=200) :: infile, mapfile, resultfile, logfile     ! path and name of the in, map, result and log file
    CHARACTER (len=100) :: fmt  ! format variable to write the output
    INTEGER nrow       ! number of lines in the datafile
    INTEGER ncol       ! number of columns in the datafile
    INTEGER nfactor    ! number of factors in the datafile
    INTEGER nrecord    ! number of data records
    CHARACTER (len=10), ALLOCATABLE :: header(:)  ! contains the names of the Markers
    REAL, ALLOCATABLE, DIMENSION(:,:) :: rawdata   ! the raw data from the data file
    REAL, ALLOCATABLE, DIMENSION(:,:) :: data   ! the data from the data file w/o the first collumn
    INTEGER, DIMENSION(100) :: mapping !mapping of the tma-data-file

    !   variables of global importance
    INTEGER n          ! total number of makers
    INTEGER k          ! number of markers to be in the reference set
    INTEGER nsample    ! number of resampling procedures
    INTEGER, ALLOCATABLE, DIMENSION(:) :: refset   ! inidices of the markers in the reference set
    INTEGER, ALLOCATABLE, DIMENSION(:) :: testset  ! inidices of the markers in the test set

    !   loop-controll variables
    INTEGER i, j, l

    !   calculated
    REAL, ALLOCATABLE, DIMENSION(:,:) :: corrmatrix    ! correlation matrix
    REAL, ALLOCATABLE, DIMENSION(:) :: ssq_best  ! minimum sum of squares of all reference factors for the best test order
    REAL, ALLOCATABLE, DIMENSION(:) :: ssq_g_data      ! all ssqs
    INTEGER, ALLOCATABLE, DIMENSION(:,:) :: order_data  ! all optimal orders

    !   output related variables
    CHARACTER (len = 300 ) record !Buffer string for output
    INTEGER logfileunit, resultfileunit     !identification ids for the output file units
    CHARACTER (len=8) :: date   !date retreved from intrinsic subroutine DATE_AND_TIME(date, time, zone, values)
    CHARACTER (len=10) :: time  !time retreved from intrinsic subroutine DATE_AND_TIME(date, time, zone, values)
    INTEGER, DIMENSION(8) :: timevector !int vector retreved from intrinsic subroutine DATE_AND_TIME(date, time, zone, values)
    CHARACTER (len=20) :: timechars(4)  !buffer vector for output of elapsed time
    DOUBLE PRECISION starttime, finishtime, elapsedtime !time stamps in seconds (year and month not included)
    LOGICAL univecfound ! triggers appropriate logging in case uniform marker vector


    !initialisation ####################################################################

    !   initialize MPI
    CALL MPI_Init(error)
    !   get the number of processes
    CALL MPI_Comm_size(MPI_COMM_WORLD, p, error)
    !   get the individual process ID
    CALL MPI_Comm_rank(MPI_COMM_WORLD, id, error)

    !   argument retrival
    numarg = iargc( )
    numarg = iargc( )
    IF(numarg < 5) THEN
        IF(id == 0) THEN
            WRITE(*,*) 'not enough arguments submitted'
            WRITE(*,*) 'scheme: infile mapfile resultfile logfile nsample (resamplingmethod) (k)'
            WRITE(*,*) 'the specification of the files is mandatory'
            WRITE(*,*) 'the specification of the number of resamplings is mandatory'
            WRITE(*,*) 'programm cancled'
        ENDIF
        STOP
    ELSE
        CALL getarg (1, infile)
        INQUIRE(FILE=infile, EXIST=file_exists)
        IF(.NOT.file_exists) THEN
            IF(id == 0) THEN
                WRITE (*,*) 'Data file not found'
                WRITE (*,*) 'Programm cancled'
            ENDIF
            STOP
        ENDIF
        CALL getarg (2, mapfile)
        INQUIRE(FILE=mapfile, EXIST=file_exists)
        IF(.NOT.file_exists) THEN
            IF(id == 0) THEN
                WRITE (*,*) 'mapping file not found'
                WRITE (*,*) 'programm cancled'
            ENDIF
            STOP
        ENDIF
        CALL getarg (3, resultfile)
        CALL getarg (4, logfile)
        CALL getarg (5, argk)
        READ(argk,*) nsample
        IF(numarg > 5) THEN
            CALL getarg (6, resamplingmethod)
            IF((resamplingmethod /= 's' .and. resamplingmethod /='S') .and. (resamplingmethod /= 'b' .and. resamplingmethod /= 'B')) THEN
                IF(id == 0) THEN
                    WRITE(*,*) 'Specified resampling method is invalid'
                    WRITE(*,*) 'Default value of s = shuffling is used'
                ENDIF
                resamplingmethod = 's'  !default resampling method: s = shuffling
            ENDIF
        ENDIF
        IF(numarg > 6) THEN
            CALL getarg (7, argk)
            READ(argk,*) k
        ELSE
            IF(id == 0) THEN
                WRITE (*,*) 'No reference set size specified'
                WRITE (*,*) 'Default value of 6 is used'
            ENDIF
            k = 6 ! default number of markers to be in the reference set
        ENDIF
        IF(numarg > 7 .AND. id == 0) THEN
            WRITE (*,*) 'Arguments after size of reference set irgnored'
        ENDIF
    ENDIF

    !   open logfile (only in master process)
    IF(id == 0) THEN
        CALL get_unit(logfileunit)
        OPEN (unit = logfileunit, file = logfile)
        CALL DATE_AND_TIME(date, time, values = timevector)
        starttime = timevector(3)*24*3600+timevector(5)*3600+timevector(6)*60+timevector(7)+DBLE(timevector(8))/DBLE(1000)
        WRITE(logfileunit,*) 'Starting job:             ',date(7:8),'.',date(5:6),'.',date(1:4),' ',time(1:2),':',time(3:4),':',time(5:10)
    ENDIF

    !Map the data file
    CALL map_tma_data_file(infile, mapfile, ncol, nrow, nfactor, nrecord, mapping, n)

    !   determine size of seeds for random number generator
    CALL RANDOM_SEED(SIZE=seedlength)

    !   allocation
    ALLOCATE(header(n)) ! allocate space for the names of the Markers
    ALLOCATE(rawdata(nrecord, n))  ! allocate space for the raw data from the data file
    ALLOCATE(data(nrecord, n))  ! allocate space for the data from the data file
    ALLOCATE(refset(k))
    ALLOCATE(testset(n-k))
    ALLOCATE(corrmatrix(n-k, k))  ! allocate space for the correlation matrix
    ALLOCATE(ssq_best(k)) ! allocate space for ssq_best
    ALLOCATE(ssq_g_data(nsample))
    ALLOCATE(order_data(nsample, n-k))
    ALLOCATE(task_distribution(p,2))
    ALLOCATE(seed(seedlength))

    !Determine the distribution of tasks
    CALL task_division(nsample, p, task_distribution)

    !intput ####################################################################
    !Read in the datafile
    CALL read_tma_data_file(infile, ncol, nrow, nfactor, nrecord, mapping, n, header, rawdata)


    !output ####################################################################
    !   logging stuff (only in master process)
    IF(id == 0) THEN
        WRITE(logfileunit,*)
        WRITE(logfileunit,'(A13, TR2, A75)') 'file:', infile
        WRITE(logfileunit,'(A13, TR2, A75)') 'mapfile:', mapfile
        WRITE(logfileunit,'(A13, TR2, I6)') 'rows:', nrow
        WRITE(logfileunit,'(A13, TR2, I6)') 'columns:', ncol
        WRITE(logfileunit,'(A13, TR2, I6)') 'factors:', nfactor
        WRITE(logfileunit,'(A13, TR2, I6)') 'n:', n
        WRITE(logfileunit,'(A13, TR2, I6)') 'reference:', k
        WRITE(logfileunit,'(A13, TR2, I6)') 'test:', n-k
        WRITE(logfileunit,'(A13, TR2, I6)') 'processes:', p
        WRITE(logfileunit,'(A13, TR2, I6)') 'resamplings:', nsample
        IF(resamplingmethod == 's' .or. resamplingmethod == 'S') THEN
            WRITE(logfileunit,'(A13, TR2, A13)') 'method:', 'shuffling'
        ENDIF
        IF(resamplingmethod == 'b' .or. resamplingmethod == 'B') THEN
            WRITE(logfileunit,'(A13, TR2, A13)') 'method:', 'bootstrapping'
        ENDIF
        WRITE(fmt,*) '(A13, TR2,', n, 'A)'
        WRITE(logfileunit,fmt) 'marker:', header(:)
        IF(resamplingmethod == 'b' .or. resamplingmethod == 'B') THEN
            header(:) = ' '
            univecfound = .FALSE.
            DO i = 1, n
                IF(SUM(rawdata(:,i)-SUM(rawdata(:,i))/nrecord) == 0) THEN
                    header(i) = 'X'
                    univecfound = .TRUE.
                ENDIF
            ENDDO
            IF(univecfound == .TRUE.) THEN
                WRITE(logfileunit,fmt) 'uniform:', header(:)
                WRITE(logfileunit,*) 'Uniform marker vector(s) found!'
                WRITE(logfileunit,*) 'The Person correlation can not be calculated correctly with this marker(s).'
                WRITE(logfileunit,*) 'Instead a correlation value near to 0 is randomly generated.'
            ENDIF
        ENDIF
        WRITE(logfileunit,*)
        WRITE(logfileunit,*) 'distribution of tasks:'
        WRITE(logfileunit,*) 'processes     first task      last task'
        DO i = 1, p
            WRITE(logfileunit,'(3X I4, 8X I4, 13X I4)') i, task_distribution(i,1), task_distribution(i,2)
        ENDDO
        WRITE(logfileunit,*)
        CALL DATE_AND_TIME(date, time)
        WRITE(logfileunit,*) 'Starting calculation:     ',date(7:8),'.',date(5:6),'.',date(1:4),' ',time(1:2),':',time(3:4),':',time(5:10)


    !reporting of the cor table and the coefficients of the regression lines for the first row
!        data = rawdata
!        DO i = 1, k
!            refset(i) = i
!        ENDDO
!        testset = 0
!        DO i = 1, n-k
!            testset(i) = i+k
!        ENDDO
!        DO i = 1, n-k
!            DO j = 1, k
!                CALL pearsoncorrelation(nrecord, data(:,testset(i)), data(:,refset(j)), corrmatrix(i,j))
!            ENDDO
!        ENDDO
!        CALL perm_tj_reporting(resultfile, corrmatrix, k, n-k)
    ENDIF


    !calculation ###############################################################
    seed(:) =  task_distribution(id+1,1)
    CALL RANDOM_SEED (PUT = seed(:))
    DO l = task_distribution(id+1,1), task_distribution(id+1,2)
        data = rawdata
        IF((resamplingmethod == 's' .or. resamplingmethod == 'S') .and. l > 1) CALL shuffle(data, nrecord, n)
        IF((resamplingmethod == 'b' .or. resamplingmethod == 'B') .and. l > 1) CALL bootstrap(data, nrecord, n, 1)
        DO i = 1, k
            refset(i) = i
        ENDDO
        testset = 0
        DO i = 1, n-k
            testset(i) = i+k
        ENDDO
        DO i = 1, n-k
            DO j = 1, k
                CALL pearsoncorrelation(nrecord, data(:,testset(i)), data(:,refset(j)), corrmatrix(i,j))
            ENDDO
        ENDDO
        CALL perm_tj(corrmatrix, k, n-k, ssq_best, ssq_g_data(l), order_data(l,:))
    ENDDO

    !log time calculations finishes
    IF(id == 0) THEN
        CALL DATE_AND_TIME(date, time, values = timevector)
        WRITE(logfileunit,*) 'Finished calculation:     ',date(7:8),'.',date(5:6),'.',date(1:4),' ',time(1:2),':',time(3:4),':',time(5:10)
    ENDIF


    !collect data ##############################################################
    tag = 1
    IF(id == 0) THEN
        DO i = 2, p
            DO j = task_distribution(i,1), task_distribution(i,2)
                !recive
                CALL MPI_Recv(ssq_g_data(j), 1, MPI_REAL, i-1, tag, MPI_COMM_WORLD, status, error)
                tag = tag + 1
                CALL MPI_Recv(order_data(j,:), n-k, MPI_INTEGER, i-1, tag, MPI_COMM_WORLD, status, error)
                tag = tag + 1
            END DO
        END DO
    ELSE
        DO i = 2, p
            DO j = task_distribution(i,1), task_distribution(i,2)
                !send
                IF(id == i-1) THEN
                    CALL MPI_Send(ssq_g_data(j), 1, MPI_REAL, 0, tag , MPI_COMM_WORLD, error)
                ENDIF
                tag = tag + 1
                IF(id == i-1) THEN
                    CALL MPI_Send(order_data(j,:), n-k, MPI_INTEGER, 0, tag , MPI_COMM_WORLD, error)
                ENDIF
                tag = tag + 1
            END DO
        ENDDO
    ENDIF

    !close mpi
    call MPI_Finalize(error)


    !output (only executed in master process) ##################################
    IF(id == 0) THEN
        !write results to file
        CALL get_unit(resultfileunit)
        OPEN (unit = resultfileunit, file = resultfile)
        DO i = 1, nsample
        record = ' '
            CALL csv_record_append_re (ssq_g_data(i), record)
            DO j = 1, n-k
                CALL  csv_record_append_i4 (order_data(i,j), record)
            ENDDO
        WRITE(resultfileunit,'(a)') trim(record)
        ENDDO
        !Log elapsed time
        CALL DATE_AND_TIME(date, time, values = timevector)
        WRITE(logfileunit,*) 'Finished collecting data: ',date(7:8),'.',date(5:6),'.',date(1:4),' ',time(1:2),':',time(3:4),':',time(5:10)
        ! calculate elapsed time
        finishtime = timevector(3)*24*3600+timevector(5)*3600+timevector(6)*60+timevector(7)+DBLE(timevector(8))/DBLE(1000)
        elapsedtime = finishtime-starttime
        ! days
        WRITE (timechars(1),*) INT(elapsedtime/(24*3600))
        timechars(1) = adjustl(timechars(1))
        ! hours
        WRITE (timechars(2),*) INT(mod(elapsedtime,DBLE(24*3600))/3600)
        timechars(2) = adjustl(timechars(2))
        ! minutes
        WRITE (timechars(3),*) INT(mod(mod(elapsedtime,DBLE(24*3600)),DBLE(3600))/60)
        timechars(3) = adjustl(timechars(3))
        ! seconds
        WRITE (timechars(4),'(F6.3)') mod(mod(mod(elapsedtime,DBLE(24*3600)),DBLE(3600)),DBLE(60))
        timechars(4) = adjustl(timechars(4))
        ! write elapsed time to file
        WRITE(logfileunit,*)
        IF (timechars(1) /= '0') THEN
            WRITE(logfileunit,*) 'Finished job after '//trim(timechars(1))//' days, '//trim(timechars(2))//' hours, '//trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
        ELSE
            IF (timechars(2) /= '0') THEN
                WRITE(logfileunit,*) 'Finished job after '//trim(timechars(2))//' hours, '//trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
            ELSE
                IF (timechars(3) /= '0') THEN
                    WRITE(logfileunit,*) 'Finished job after '//trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
                ELSE
                    WRITE(logfileunit,*) 'Finished job after '//trim(timechars(4))//' seconds'
                ENDIF
            ENDIF
        ENDIF
        !close log file
        CLOSE(logfileunit)
        !report to console for performance tests
        WRITE(record,*) p
        record = ADJUSTL(record)
        IF (timechars(1) /= '0') THEN
            IF(p == 1) THEN
                WRITE(*,*) 'Finished job on '//trim(record)//' process after '//trim(timechars(1))//' days, '//trim(timechars(2))//' hours, '//trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
            ELSE
                WRITE(*,*) 'Finished job on '//trim(record)//' processes after '//trim(timechars(1))//' days, '//trim(timechars(2))//' hours, '//trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
            ENDIF
        ELSE
            IF (timechars(2) /= '0') THEN
                IF(p == 1) THEN
                    WRITE(*,*) 'Finished job on '//trim(record)//' process after '//trim(timechars(2))//' hours, '//trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
                ELSE
                    WRITE(*,*) 'Finished job on '//trim(record)//' processes after '//trim(timechars(2))//' hours, '//trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
                ENDIF
            ELSE
                IF (timechars(3) /= '0') THEN
                    IF(p == 1) THEN
                        WRITE(*,*) 'Finished job on '//trim(record)//' process after '//trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
                    ELSE
                        WRITE(*,*) 'Finished job on '//trim(record)//' processes after '//trim(timechars(3))//' minutes and '//trim(timechars(4))//' seconds'
                    ENDIF
                ELSE
                    IF(p == 1) THEN
                        WRITE(*,*) 'Finished job on '//trim(record)//' process after '//trim(timechars(4))//' seconds'
                    ELSE
                        WRITE(*,*) 'Finished job on '//trim(record)//' processes after '//trim(timechars(4))//' seconds'
                    ENDIF
                ENDIF
            ENDIF
        ENDIF
    ENDIF

END PROGRAM perm2011f90_resample_mpi
